"""Odoo Series and Editions."""

from enum import Enum
from typing import Optional, Set

from .addons_set import AddonsSet
from .exceptions import UnsupportedOdooSeries

__all__ = [
    "OdooEdition",
    "OdooSeries",
    "detect_from_addon_version",
    "detect_from_addons_set",
]


MIN_VERSION_PARTS = 5


class OdooSeries(str, Enum):
    """Enum representing an Odoo Series (also known as Version)."""

    v8_0 = "8.0"
    v9_0 = "9.0"
    v10_0 = "10.0"
    v11_0 = "11.0"
    v12_0 = "12.0"
    v13_0 = "13.0"
    v14_0 = "14.0"
    v15_0 = "15.0"
    v16_0 = "16.0"
    v17_0 = "17.0"
    v18_0 = "18.0"

    @classmethod
    def from_str(cls, value: str, context: Optional[str] = None) -> "OdooSeries":
        """Get the OdooSeries from a string.

        Raise UnsupportedOdooSeries if the string is not recognized.
        """
        try:
            return cls(value)
        except ValueError as e:
            msg = f"Unsupported Odoo Series: {value}"
            if context:
                msg = f"{msg} in {context}"
            raise UnsupportedOdooSeries(msg) from e


class OdooEdition(str, Enum):
    """Enum representing an Odoo Edition."""

    CE = "c"
    EE = "e"


def detect_from_addon_version(version: str) -> Optional[OdooSeries]:
    """Detect the Odoo Series from an addon version.

    Returns ``None`` if the version is not recognized.
    """
    parts = version.split(".")
    if len(parts) < MIN_VERSION_PARTS:
        return None
    try:
        return OdooSeries(".".join(parts[:2]))
    except ValueError:
        return None


def detect_from_addons_set(addons_set: AddonsSet) -> Set[OdooSeries]:
    detected: Set[OdooSeries] = set()
    for addon in addons_set.values():
        addon_version = addon.manifest.version
        if not addon_version:
            continue
        addon_series = detect_from_addon_version(addon_version)
        if not addon_series:
            continue
        detected.add(addon_series)
    return detected



import os
import subprocess
import sys
from pathlib import Path
from typing import Iterator, List, Optional, TextIO

from packaging.version import parse as parse_version

if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final

from .addon import Addon
from .exceptions import InvalidManifest, UnknownPostVersionStrategy
from .manifest import MANIFEST_NAMES, Manifest

POST_VERSION_STRATEGY_NONE: Final = "none"
POST_VERSION_STRATEGY_NINETYNINE_DEVN: Final = ".99.devN"
POST_VERSION_STRATEGY_P1_DEVN: Final = "+1.devN"
POST_VERSION_STRATEGY_DOT_N: Final = ".N"


def _run_git_command_exit_code(
    args: List[str],
    cwd: Optional[Path] = None,
    stderr: Optional[TextIO] = None,
) -> int:
    return subprocess.call(  # noqa: S603
        ["git", *args],  # noqa: S607
        cwd=cwd,
        stderr=stderr,
    )


def _run_git_command_bytes(
    args: List[str],
    cwd: Optional[Path] = None,
    stderr: Optional[TextIO] = None,
) -> str:
    output = subprocess.check_output(  # noqa: S603
        ["git", *args],  # noqa: S607
        cwd=cwd,
        universal_newlines=True,
        stderr=stderr,
    )
    return output.strip()


def _run_git_command_lines(
    args: List[str],
    cwd: Optional[Path] = None,
    stderr: Optional[TextIO] = None,
) -> List[str]:
    output = _run_git_command_bytes(args, cwd=cwd, stderr=stderr)
    return output.split("\n")


def _is_git_controlled(path: Path) -> bool:
    with Path(os.devnull).open("w") as devnull:
        r = _run_git_command_exit_code(["rev-parse"], cwd=path, stderr=devnull)
        return r == 0


def _get_git_uncommitted(path: Path) -> bool:
    r = _run_git_command_exit_code(["diff", "--quiet", "--exit-code", "."], cwd=path)
    return r != 0


def _get_git_root(path: Path) -> Path:
    return Path(_run_git_command_bytes(["rev-parse", "--show-toplevel"], cwd=path))


def _git_log_iterator(path: Path) -> Iterator[str]:
    """yield commits using git log -- <dir>"""
    n = 10
    count = 0
    while True:
        lines = _run_git_command_lines(
            ["log", "--oneline", "-n", str(n), "--skip", str(count), "--", "."],
            cwd=path,
        )
        for line in lines:
            sha = line.split(" ", 1)[0]
            count += 1
            yield sha
        if len(lines) < n:
            break


def _read_manifest_from_sha(
    sha: str,
    addon_dir: Path,
    git_root: Path,
) -> Optional[Manifest]:
    rel_addon_dir = addon_dir.relative_to(git_root)
    for manifest_name in MANIFEST_NAMES:
        manifest_path = rel_addon_dir / manifest_name
        try:
            with Path(os.devnull).open("w") as devnull:
                s = _run_git_command_bytes(
                    ["show", f"{sha}:{manifest_path}"],
                    cwd=git_root,
                    stderr=devnull,
                )
        except subprocess.CalledProcessError:
            continue
        try:
            return Manifest.from_str(s)
        except InvalidManifest:
            break
    return None


def _bump_last(version: str) -> str:
    int_version = [int(i) for i in version.split(".")]
    int_version[-1] += 1
    return ".".join(str(i) for i in int_version)


def get_git_postversion(  # noqa: C901, PLR0911, PLR0912 too complex
    addon: Addon,
    strategy: str,
) -> str:
    """return the addon version number, with a developmental version increment
    if there were git commits in the addon_dir after the last version change.

    If the last change to the addon correspond to the version number in the
    manifest it is used as is for the python package version. Otherwise a
    counter is incremented for each commit and resulting version number has
    the following form, depending on the strategy (N being the number of git
    commits since the version change):

    * STRATEGY_NONE: return the version in the manifest as is
    * STRATEGY_99_DEVN: [8|9].0.x.y.z.99.devN
    * STRATEGY_P1_DEVN: [series].0.x.y.(z+1).devN
    * STRATEGY_DOT_N: [series].0.x.y.z.N

    Notes:

    * pip ignores .postN  by design (https://github.com/pypa/pip/issues/2872)
    * x.y.z.devN is anterior to x.y.z

    Note: we don't put the sha1 of the commit in the version number because
    this is not PEP 440 compliant and is therefore misinterpreted by pip.
    """
    last_version = addon.manifest.version or "0.0.0"
    addon_dir = addon.path.resolve()
    if strategy == POST_VERSION_STRATEGY_NONE:
        return last_version
    last_version_parsed = parse_version(last_version)
    if not _is_git_controlled(addon_dir):
        return last_version
    if _get_git_uncommitted(addon_dir):
        uncommitted = True
        count = 1
    else:
        uncommitted = False
        count = 0
    last_sha = None
    git_root = _get_git_root(addon_dir)
    for sha in _git_log_iterator(addon_dir):
        manifest = _read_manifest_from_sha(sha, addon_dir, git_root)
        if manifest is None:
            break
        version = manifest.version or "0.0.0"
        version_parsed = parse_version(version)
        if version_parsed != last_version_parsed:
            break
        if last_sha is None:
            last_sha = sha
        else:
            count += 1
    if not count:
        return last_version
    if last_sha:
        if strategy == POST_VERSION_STRATEGY_NINETYNINE_DEVN:
            return last_version + f".99.dev{count}"
        if strategy == POST_VERSION_STRATEGY_P1_DEVN:
            return _bump_last(last_version) + f".dev{count}"
        if strategy == POST_VERSION_STRATEGY_DOT_N:
            return last_version + f".{count}"
        msg = f"Unknown postversion strategy: {strategy}"
        raise UnknownPostVersionStrategy(msg)
    if uncommitted:
        return last_version + ".dev1"
    # if everything is committed, the last commit
    # must have the same version as current,
    # so last_sha must be set and we'll never reach this branch
    return last_version



"""Information about core Odoo addons."""

import sys
from functools import lru_cache
from typing import Set

if sys.version_info >= (3, 9):
    from importlib.resources import files as package_files
else:
    from importlib_resources import files as package_files

from ..odoo_series import OdooEdition, OdooSeries

__all__ = [
    "get_core_addon_license",
    "get_core_addons",
    "is_core_addon",
    "is_core_ce_addon",
    "is_core_ee_addon",
]


@lru_cache()
def _get_core_addons(odoo_series: OdooSeries, odoo_edition: OdooEdition) -> Set[str]:
    with package_files("manifestoo_core.core_addons").joinpath(
        f"addons-{odoo_series.value}-{odoo_edition.value}.txt",
    ).open() as f:
        return {a.strip() for a in f if not a.startswith("#")}


@lru_cache()
def get_core_addons(odoo_series: OdooSeries) -> Set[str]:
    """Return the set of core Odoo addons for a given Odoo series."""
    return _get_core_addons(odoo_series, OdooEdition.CE) | _get_core_addons(
        odoo_series,
        OdooEdition.EE,
    )


def is_core_ce_addon(addon_name: str, odoo_series: OdooSeries) -> bool:
    """Test wether an addon is part of the Community Edition of a given Odoo series."""
    return addon_name in _get_core_addons(odoo_series, OdooEdition.CE)


def is_core_ee_addon(addon_name: str, odoo_series: OdooSeries) -> bool:
    """Test wether an addon is part of the Enterprise Edition of a given Odoo series."""
    return addon_name in _get_core_addons(odoo_series, OdooEdition.EE)


def is_core_addon(addon_name: str, odoo_series: OdooSeries) -> bool:
    """Test wether an addon is part of a given Odoo series."""
    return addon_name in get_core_addons(odoo_series)


def get_core_addon_license(addon_name: str, odoo_series: OdooSeries) -> str:
    """Get the license of a core Odoo addon.

    This function overrides any license set in the upstream addon
    manifest, as Odoo has a uniform license for each version and
    edition, and the manifests have been known to be unreliable in that
    respect.
    """
    if is_core_ce_addon(addon_name, odoo_series):
        if odoo_series == OdooSeries.v8_0:
            return "AGPL-3"
        return "LGPL-3"
    if is_core_ee_addon(addon_name, odoo_series):
        return "OEEL-1"
    msg = f"{addon_name} is not a core addon."  # pragma: no cover
    raise AssertionError(msg)  # pragma: no cover



from pathlib import Path
from typing import Any, Dict

from manifestoo_core.addon import Addon
from manifestoo_core.addons_set import AddonsSet
from manifestoo_core.manifest import Manifest


def populate_addons_dir(addons_dir: Path, addons: Dict[str, Dict[str, Any]]) -> None:
    if not addons_dir.is_dir():
        addons_dir.mkdir()
    for addon_name, manifest in addons.items():
        addon_path = addons_dir / addon_name
        addon_path.mkdir()
        (addon_path / "__init__.py").touch()
        manifest_path = addon_path / "__manifest__.py"
        manifest_path.write_text(repr(manifest))


def mock_addons_set(addons: Dict[str, Dict[str, Any]]) -> AddonsSet:
    addons_set = AddonsSet()
    for addon_name, manifest_dict in addons.items():
        manifest = Manifest.from_dict(manifest_dict)
        manifest_path = (
            Path("/tmp/fake-addons-dir") / addon_name / "__manifest__.py"  # noqa: S108
        )
        addons_set[addon_name] = Addon(manifest, manifest_path)
    return addons_set



from pathlib import Path
from typing import Optional

from .exceptions import (
    AddonNotFound,
    AddonNotFoundInvalidManifest,
    AddonNotFoundNoInit,
    AddonNotFoundNoManifest,
    AddonNotFoundNotADirectory,
    AddonNotFoundNotInstallable,
)
from .manifest import InvalidManifest, Manifest, get_manifest_path

__all__ = [
    "Addon",
    "is_addon_dir",
]


def is_addon_dir(addon_dir: Path, allow_not_installable: bool = False) -> bool:
    """Detect if a directory contains an Odoo addon.

    :param addon_dir: The directory to check.
    :param allow_not_installable: Whether to allow the addon to be have
        installable=False in its manifest.
    """
    try:
        Addon.from_addon_dir(addon_dir, allow_not_installable)
    except AddonNotFound:
        return False
    else:
        return True


class Addon:
    """Represent a concrete addon manifest."""

    manifest: Manifest
    "The addon's manifest."
    manifest_path: Path
    "The path to the addon's manifest file."
    name: str
    "The addon's name."

    def __init__(
        self,
        manifest: Manifest,
        manifest_path: Path,
        name: Optional[str] = None,
    ) -> None:
        """Do not use this constructor, use the from_addon_dir classmethod instead."""
        self.manifest = manifest
        self.manifest_path = manifest_path
        self.path = self.manifest_path.parent
        if name is None:
            self.name = self.path.name
        else:
            self.name = name

    @classmethod
    def from_addon_dir(
        cls,
        addon_dir: Path,
        allow_not_installable: bool = False,
    ) -> "Addon":
        """Obtain an Addon object from an addon directory path.

        Raises an :class:`AddonNotFound` exception if the directory is not a valid addon
        directory, or if ``allow_not_installable`` is False and the addon is not
        installable.
        """
        if not addon_dir.is_dir():
            msg = f"{addon_dir} is not a directory"
            raise AddonNotFoundNotADirectory(msg)
        manifest_path = get_manifest_path(addon_dir)
        if not manifest_path:
            msg = f"No manifest file found in {addon_dir}"
            raise AddonNotFoundNoManifest(msg)
        try:
            manifest = Manifest.from_file(manifest_path)
            if not allow_not_installable and not manifest.installable:
                msg = f"{addon_dir} is not installable"
                raise AddonNotFoundNotInstallable(msg)
        except InvalidManifest as e:
            raise AddonNotFoundInvalidManifest(str(e)) from e
        if not addon_dir.joinpath("__init__.py").is_file():
            msg = f"{addon_dir} is missing an __init__.py"
            raise AddonNotFoundNoInit(msg)
        return cls(manifest, manifest_path)



import email.parser
import re
import sys
import warnings
from dataclasses import dataclass
from email.message import Message
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Sequence, Set, Tuple, Union

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from .addon import Addon
from .core_addons import get_core_addons
from .exceptions import (
    InvalidDistributionName,
    UnsupportedManifestVersion,
    UnsupportedOdooSeries,
)
from .git_postversion import (
    POST_VERSION_STRATEGY_DOT_N,
    POST_VERSION_STRATEGY_NINETYNINE_DEVN,
    POST_VERSION_STRATEGY_NONE,
    POST_VERSION_STRATEGY_P1_DEVN,
    get_git_postversion,
)
from .manifest import Manifest
from .odoo_series import MIN_VERSION_PARTS, OdooSeries

ODOO_ADDON_DIST_RE = re.compile(
    r"^(odoo(\d{1,2})?[-_]addon[-_].*|odoo$|odoo[^a-zA-Z0-9._-]+)",
    re.IGNORECASE,
)
ODOO_ADDON_METADATA_NAME_RE = re.compile(
    r"^odoo(\d{1,2})?[-_]addon[-_](?P<addon_name>[a-zA-Z0-9_-]+)$",
)

__all__ = [
    "POST_VERSION_STRATEGY_DOT_N",
    "POST_VERSION_STRATEGY_NINETYNINE_DEVN",
    "POST_VERSION_STRATEGY_NONE",
    "POST_VERSION_STRATEGY_P1_DEVN",
    "addon_name_to_distribution_name",
    "addon_name_to_requirement",
    "distribution_name_to_addon_name",
    "metadata_from_addon_dir",
]


def _filter_odoo_addon_dependencies(dependencies: Sequence[str]) -> Iterator[str]:
    """Filter a sequence of python packages dependencies to exclude 'odoo' and odoo
    addons."""
    for dependency in dependencies:
        if dependency.startswith("odoo>="):
            continue
        if ODOO_ADDON_DIST_RE.match(dependency):
            continue
        yield dependency


class MetadataOptions(TypedDict, total=False):
    """A dictionary of options for metadata conversion.

    Supported keys are:

    - ``depends_override``
    - ``external_dependencies_override``
    - ``external_dependencies_only``
    - ``odoo_series_override`` and ``odoo_version_override``
    - ``post_version_strategy_override``
    """

    depends_override: Optional[Dict[str, str]]
    external_dependencies_override: Optional[
        Dict[str, Dict[str, Union[str, List[str]]]]
    ]
    external_dependencies_only: Optional[bool]
    odoo_series_override: Optional[str]
    odoo_version_override: Optional[str]
    post_version_strategy_override: Optional[str]


def metadata_from_addon_dir(
    addon_dir: Path,
    options: Optional[MetadataOptions] = None,
    precomputed_metadata_file: Optional[Path] = None,
) -> Message:
    """Return Python Package Metadata 2.1 for an Odoo addon directory as an
    ``email.message.Message``.

    The Description field is absent and is stored in the message payload. All values are
    guaranteed to not contain newline characters, except for the payload.

    ``precomputed_metadata_path`` may point to a file containing pre-computed metadata
    that will be used to obtain the Name and Version, instead of looking at the addon
    directory name or manifest version + VCS, respectively. This is useful to process a
    manifest from a sdist tarball with PKG-INFO, for example, when the original
    directory name or VCS is not available to compute the package name and version.

    This function may raise :class:`manifestoo_core.exceptions.ManifestooException` if
    ``addon_dir`` does not contain a valid installable Odoo addon for a supported Odoo
    version.
    """
    if options is None:
        options = MetadataOptions()
    addon = Addon.from_addon_dir(addon_dir)
    manifest = addon.manifest

    if precomputed_metadata_file and precomputed_metadata_file.is_file():
        with precomputed_metadata_file.open(encoding="utf-8") as fp:
            pkg_info = email.parser.HeaderParser().parse(fp)
            addon_name = distribution_name_to_addon_name(pkg_info["Name"])
            version = pkg_info["Version"]
        _, odoo_series, odoo_series_info = _get_version(
            addon,
            options.get("odoo_series_override") or options.get("odoo_version_override"),
            git_post_version=False,
        )
    else:
        addon_name = addon_dir.absolute().name
        version, odoo_series, odoo_series_info = _get_version(
            addon,
            options.get("odoo_series_override") or options.get("odoo_version_override"),
            git_post_version=True,
            post_version_strategy_override=options.get(
                "post_version_strategy_override",
            ),
        )
    install_requires = _get_install_requires(
        odoo_series_info,
        manifest,
        depends_override=options.get("depends_override"),
        external_dependencies_override=options.get("external_dependencies_override"),
    )
    # Update Requires-Dist metadata with dependencies that are not odoo nor odoo addons
    if options.get("external_dependencies_only"):
        install_requires = list(_filter_odoo_addon_dependencies(install_requires))

    def _set(key: str, value: Union[None, str, List[str]]) -> None:
        if not value:
            return
        if isinstance(value, list):
            for v in value:
                meta[key] = v
        else:
            meta[key] = value

    meta = Message()
    _set("Metadata-Version", "2.1")
    _set("Name", _addon_name_to_metadata_name(addon_name, odoo_series_info))
    _set("Version", version)
    _set("Requires-Python", odoo_series_info.python_requires)
    _set("Requires-Dist", install_requires)
    _set("Summary", _no_nl(manifest.summary or manifest.name))
    _set("Home-page", manifest.website)
    _set("License", manifest.license)
    _set("Author", _no_nl(manifest.author))
    _set("Author-email", _author_email(manifest.author))
    _set("Classifier", _make_classifiers(odoo_series, manifest))
    long_description = _long_description(addon)
    if long_description:
        meta.set_payload(long_description)

    return meta


@dataclass
class OdooSeriesInfo:
    odoo_dep: str
    pkg_name_pfx: str
    python_requires: str
    universal_wheel: bool
    git_postversion_strategy: str
    core_addons: Set[str]
    pkg_version_specifier: str = ""
    addons_ns: Optional[str] = None
    namespace_packages: Optional[List[str]] = None

    @classmethod
    def from_odoo_series(
        cls,
        odoo_series: OdooSeries,
        context: Optional[str] = None,
    ) -> "OdooSeriesInfo":
        try:
            return ODOO_SERIES_INFO[odoo_series]
        except KeyError as e:
            msg = "Unsupported odoo series '{odoo_series_str}'"
            if context:
                msg = f"{msg} in {context}"
            raise UnsupportedOdooSeries(msg) from e


ODOO_SERIES_INFO = {
    OdooSeries.v8_0: OdooSeriesInfo(
        odoo_dep="odoo>=8.0a,<9.0a",
        pkg_name_pfx="odoo8-addon",
        addons_ns="odoo_addons",
        namespace_packages=["odoo_addons"],
        python_requires="~=2.7",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        core_addons=get_core_addons(OdooSeries.v8_0),
    ),
    OdooSeries.v9_0: OdooSeriesInfo(
        odoo_dep="odoo>=9.0a,<9.1a",
        pkg_name_pfx="odoo9-addon",
        addons_ns="odoo_addons",
        namespace_packages=["odoo_addons"],
        python_requires="~=2.7",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        core_addons=get_core_addons(OdooSeries.v9_0),
    ),
    OdooSeries.v10_0: OdooSeriesInfo(
        odoo_dep="odoo>=10.0,<10.1dev",
        pkg_name_pfx="odoo10-addon",
        addons_ns="odoo.addons",
        namespace_packages=["odoo", "odoo.addons"],
        python_requires="~=2.7",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        core_addons=get_core_addons(OdooSeries.v10_0),
    ),
    OdooSeries.v11_0: OdooSeriesInfo(
        odoo_dep="odoo>=11.0a,<11.1dev",
        pkg_name_pfx="odoo11-addon",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
        universal_wheel=True,
        git_postversion_strategy=POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        core_addons=get_core_addons(OdooSeries.v11_0),
    ),
    OdooSeries.v12_0: OdooSeriesInfo(
        odoo_dep="odoo>=12.0a,<12.1dev",
        pkg_name_pfx="odoo12-addon",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.5",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_NINETYNINE_DEVN,
        core_addons=get_core_addons(OdooSeries.v12_0),
    ),
    OdooSeries.v13_0: OdooSeriesInfo(
        odoo_dep="odoo>=13.0a,<13.1dev",
        pkg_name_pfx="odoo13-addon",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.5",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_P1_DEVN,
        core_addons=get_core_addons(OdooSeries.v13_0),
    ),
    OdooSeries.v14_0: OdooSeriesInfo(
        odoo_dep="odoo>=14.0a,<14.1dev",
        pkg_name_pfx="odoo14-addon",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.6",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_P1_DEVN,
        core_addons=get_core_addons(OdooSeries.v14_0),
    ),
    OdooSeries.v15_0: OdooSeriesInfo(
        odoo_dep="odoo>=15.0a,<15.1dev",
        pkg_name_pfx="odoo-addon",
        pkg_version_specifier=">=15.0dev,<15.1dev",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.8",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_DOT_N,
        core_addons=get_core_addons(OdooSeries.v15_0),
    ),
    OdooSeries.v16_0: OdooSeriesInfo(
        odoo_dep="odoo>=16.0a,<16.1dev",
        pkg_name_pfx="odoo-addon",
        pkg_version_specifier=">=16.0dev,<16.1dev",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.10",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_DOT_N,
        core_addons=get_core_addons(OdooSeries.v16_0),
    ),
    OdooSeries.v17_0: OdooSeriesInfo(
        odoo_dep="odoo>=17.0a,<17.1dev",
        pkg_name_pfx="odoo-addon",
        pkg_version_specifier=">=17.0dev,<17.1dev",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.10",
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_DOT_N,
        core_addons=get_core_addons(OdooSeries.v17_0),
    ),
    OdooSeries.v18_0: OdooSeriesInfo(
        odoo_dep="odoo==18.0.*",
        pkg_name_pfx="odoo-addon",
        pkg_version_specifier="==18.0.*",
        addons_ns="odoo.addons",
        namespace_packages=None,
        python_requires=">=3.10",  # TODO To be confirmed
        universal_wheel=False,
        git_postversion_strategy=POST_VERSION_STRATEGY_DOT_N,
        core_addons=get_core_addons(OdooSeries.v18_0),
    ),
}

# map names of common python external dependencies in Odoo manifest files
# to actual python package names
EXTERNAL_DEPENDENCIES_MAP = {
    "Asterisk": "py-Asterisk",
    "coda": "pycoda",
    "cups": "pycups",
    "dateutil": "python-dateutil",
    "ldap": "python-ldap",
    "serial": "pyserial",
    "suds": "suds-jurko",
    "stdnum": "python-stdnum",
    "Crypto.Cipher.DES3": "pycrypto",
    "OpenSSL": "pyOpenSSL",
}


def distribution_name_to_addon_name(metadata_name: str) -> str:
    """Convert a packaging distribution name to the corresponding Odoo addon name."""
    mo = ODOO_ADDON_METADATA_NAME_RE.match(metadata_name)
    if not mo:
        msg = f"{metadata_name} does not look like an Odoo addon package name"
        raise InvalidDistributionName(msg)
    return mo.group("addon_name").replace("-", "_")


def _addon_name_to_metadata_name(
    addon_name: str,
    odoo_series_info: OdooSeriesInfo,
) -> str:
    return odoo_series_info.pkg_name_pfx + "-" + addon_name


def addon_name_to_distribution_name(addon_name: str, odoo_series: OdooSeries) -> str:
    """Convert an Odoo addon name to the corresponding packaging distribution name."""
    return _addon_name_to_metadata_name(
        addon_name,
        OdooSeriesInfo.from_odoo_series(odoo_series),
    )


def _addon_name_to_requires_dist(
    addon_name: str,
    odoo_series_info: OdooSeriesInfo,
) -> str:
    pkg_name = _addon_name_to_metadata_name(addon_name, odoo_series_info)
    pkg_version_specifier = odoo_series_info.pkg_version_specifier
    return pkg_name + pkg_version_specifier


def addon_name_to_requirement(addon_name: str, odoo_series: OdooSeries) -> str:
    """Convert an Odoo addon name to a requirement specifier."""
    return _addon_name_to_requires_dist(
        addon_name,
        OdooSeriesInfo.from_odoo_series(odoo_series),
    )


def _author_email(author: Optional[str]) -> Optional[str]:
    if author and "Odoo Community Association (OCA)" in author:
        return "support@odoo-community.org"
    return None


def _no_nl(s: Optional[str]) -> Optional[str]:
    if not s:
        return s
    return " ".join(s.split())


def _long_description(addon: Addon) -> Optional[str]:
    readme_path = addon.path / "README.rst"
    if readme_path.is_file():
        return readme_path.read_text(encoding="utf-8")
    return addon.manifest.description


def _make_classifiers(odoo_series: OdooSeries, manifest: Manifest) -> List[str]:
    classifiers = [
        "Programming Language :: Python",
        "Framework :: Odoo",
        f"Framework :: Odoo :: {odoo_series.value}",
    ]

    # commonly used licenses in OCA
    licenses = {
        "agpl-3": "License :: OSI Approved :: GNU Affero General Public License v3",
        "agpl-3 or any later version": (
            "License :: OSI Approved :: "
            "GNU Affero General Public License v3 or later (AGPLv3+)"
        ),
        "gpl-2": "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "gpl-2 or any later version": (
            "License :: OSI Approved :: "
            "GNU General Public License v2 or later (GPLv2+)"
        ),
        "gpl-3": "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "gpl-3 or any later version": (
            "License :: OSI Approved :: "
            "GNU General Public License v3 or later (GPLv3+)"
        ),
        "lgpl-2": (
            "License :: OSI Approved :: "
            "GNU Lesser General Public License v2 (LGPLv2)"
        ),
        "lgpl-2 or any later version": (
            "License :: OSI Approved :: "
            "GNU Lesser General Public License v2 or later (LGPLv2+)"
        ),
        "lgpl-3": (
            "License :: OSI Approved :: "
            "GNU Lesser General Public License v3 (LGPLv3)"
        ),
        "lgpl-3 or any later version": (
            "License :: OSI Approved :: "
            "GNU Lesser General Public License v3 or later (LGPLv3+)"
        ),
    }
    license = manifest.license  # `license` is shadowing a python builtin
    if license:
        license_classifier = licenses.get(license.lower())
        if license_classifier:
            classifiers.append(license_classifier)

    # commonly used development status in OCA
    development_statuses = {
        "alpha": "Development Status :: 3 - Alpha",
        "beta": "Development Status :: 4 - Beta",
        "production/stable": "Development Status :: 5 - Production/Stable",
        "stable": "Development Status :: 5 - Production/Stable",
        "production": "Development Status :: 5 - Production/Stable",
        "mature": "Development Status :: 6 - Mature",
    }
    development_status = manifest.development_status
    if development_status:
        development_status_classifer = development_statuses.get(
            development_status.lower(),
        )
        if development_status_classifer:
            classifiers.append(development_status_classifer)

    return classifiers


def _get_install_requires(
    odoo_series_info: OdooSeriesInfo,
    manifest: Manifest,
    no_depends: Optional[Set[str]] = None,
    depends_override: Optional[Dict[str, str]] = None,
    external_dependencies_override: Optional[
        Dict[str, Dict[str, Union[str, List[str]]]]
    ] = None,
) -> List[str]:
    install_requires = []
    # dependency on Odoo
    install_requires.append(odoo_series_info.odoo_dep)
    # dependencies on other addons (except Odoo official addons)
    for depend in manifest.depends:
        if depend in odoo_series_info.core_addons:
            continue
        if no_depends and depend in no_depends:
            continue
        if depends_override and depend in depends_override:
            install_require = depends_override[depend]
        else:
            install_require = _addon_name_to_requires_dist(depend, odoo_series_info)
        if install_require:
            install_requires.append(install_require)
    # python external_dependencies
    for dep in manifest.external_dependencies.get("python", []):
        if external_dependencies_override and dep in external_dependencies_override.get(
            "python", {}
        ):
            final_dep = external_dependencies_override.get("python", {})[dep]
        else:
            final_dep = EXTERNAL_DEPENDENCIES_MAP.get(dep, dep)
        if isinstance(final_dep, list):
            install_requires.extend(final_dep)
        else:
            install_requires.append(final_dep)
    return sorted(install_requires)


def _get_version(
    addon: Addon,
    odoo_series_override: Optional[str] = None,
    git_post_version: bool = True,
    post_version_strategy_override: Optional[str] = None,
) -> Tuple[str, OdooSeries, OdooSeriesInfo]:
    """Get addon version information from an addon directory"""
    version = addon.manifest.version
    if not version:
        msg = f"No version in manifest in {addon.path}"
        warnings.warn(msg, stacklevel=1)
        version = "0.0.0"
    if not odoo_series_override:
        version_parts = version.split(".")
        if len(version_parts) < MIN_VERSION_PARTS:
            msg = (
                f"Version in manifest must have at least "
                f"{MIN_VERSION_PARTS} components and start with "
                f"the Odoo series number (in {addon.path})"
            )
            raise UnsupportedManifestVersion(msg)
        odoo_series_str = ".".join(version_parts[:2])
    else:
        odoo_series_str = odoo_series_override
    odoo_series = OdooSeries.from_str(
        odoo_series_str,
        context=str(addon.path),
    )
    odoo_series_info = OdooSeriesInfo.from_odoo_series(
        odoo_series,
        context=str(addon.path),
    )
    if git_post_version:
        version = get_git_postversion(
            addon,
            post_version_strategy_override or odoo_series_info.git_postversion_strategy,
        )
    return version, odoo_series, odoo_series_info



from typing import Any

import pytest

from manifestoo_core.manifest import InvalidManifest, Manifest


@pytest.mark.parametrize(
    ("key", "value"),
    [
        ("name", "the name"),
        ("name", None),
        ("version", "1.0.0"),
        ("version", None),
        ("license", "GPL-3"),
        ("license", None),
        ("development_status", "Beta"),
        ("development_status", None),
        ("depends", ["a", "b"]),
        ("external_dependencies", {"python": ["httpx"]}),
        ("installable", True),
        ("installable", False),
    ],
)
def test_manifest_valid_value(key: str, value: Any) -> None:
    manifest = Manifest.from_dict({key: value})
    assert getattr(manifest, key) == value


@pytest.mark.parametrize(
    ("key", "value"),
    [
        ("name", 1),
        ("version", 1),
        ("license", ["i"]),
        ("development_status", {}),
        ("depends", 1),
        ("depends", {}),
        ("depends", None),
        ("external_dependencies", {1: {}}),
        ("external_dependencies", None),
        ("external_dependencies", [1]),
        ("installable", 1),
        ("installable", None),
    ],
)
def test_manifest_invalid_value(key: str, value: Any) -> None:
    manifest = Manifest.from_dict({key: value})
    with pytest.raises(InvalidManifest):
        getattr(manifest, key)


def test_manifest_non_str_keys() -> None:
    with pytest.raises(InvalidManifest):
        Manifest.from_dict({"name": "the name", 1: "1"})


def test_manifest_invalid_syntax() -> None:
    with pytest.raises(InvalidManifest):
        Manifest.from_str('{"name": "the name", ...}')


@pytest.mark.parametrize(
    ("key", "default"),
    [
        ("name", None),
        ("version", None),
        ("license", None),
        ("development_status", None),
        ("depends", []),
        ("external_dependencies", {}),
        ("installable", True),
        ("category", None),
    ],
)
def test_manifest_default_value(key: str, default: Any) -> None:
    manifest = Manifest.from_dict({})
    assert getattr(manifest, key) == default
